//
// integer.hpp
//
// Bill Seymour, 2024-02-02
//
// Copyright Bill Seymour 2024.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This is the header for an open-source unbounded integer implemented
// as a sign-magnitude, little-endian vector<something unsigned>.
//

#ifndef INTEGER_HPP_INCLUDED
#define INTEGER_HPP_INCLUDED

#include <string>
#include <vector>
#include <limits>
#include <utility>
#include <type_traits>
#include <stdexcept>

#include <cstddef>
#include <climits>
#include <cmath> // isfinite, fabs

#include "bignum_common.hpp"

namespace bignum {

class integer final
{
    //
    // We want a value_type for the vector that's as big as possible
    // while still having an extended type that's big enough to hold
    // carries and partial products.
    //
    #if ULLONG_MAX / ULONG_MAX > ULONG_MAX
      // probably int and long are both 32 bits
      using val_type = unsigned long;
      using ext_type = unsigned long long;
    #elif ULONG_MAX / UINT_MAX > UINT_MAX
      // but maybe long and long long are both 64 bits
      using val_type = unsigned int;
      using ext_type = unsigned long;
    #elif ULLONG_MAX / UINT_MAX > UINT_MAX
      // something weird like 32/48/64 bits?
      using val_type = unsigned int;
      using ext_type = unsigned long long;
    // else try almost anything
    #elif UINT_MAX / USHRT_MAX > USHRT_MAX
      using val_type = unsigned short;
      using ext_type = unsigned int;
    #elif ULONG_MAX / USHRT_MAX > USHRT_MAX
      using val_type = unsigned short;
      using ext_type = unsigned long;
    #elif UINT_MAX / UCHAR_MAX > UCHAR_MAX
      using val_type = unsigned char;
      using ext_type = unsigned int;
    #else
      #error Can't find a type for the vector's value_type.
    #endif

    static constexpr unsigned val_bits = CHAR_BIT * sizeof(val_type);
    static constexpr ext_type val_mask = static_cast<val_type>(~0UL);
    static constexpr ext_type val_mod  = val_mask + 1;

    using absval = std::vector<val_type>;
    using iter   = absval::iterator;
    using citer  = absval::const_iterator;
    using riter  = absval::reverse_iterator;
    using criter = absval::const_reverse_iterator;

    absval val;
    bool neg;

    static void pop_leading_zeros(absval& v) noexcept
    {
        // NB: little-endian
        while (!v.empty() && 0 == v.back())
        {
            v.pop_back();
        }
    }
    void disallow_neg_zero() noexcept
    {
        if (val.empty())
        {
            neg = false;
        }
    }
    void normalize() noexcept
    {
        pop_leading_zeros(val);
        disallow_neg_zero();
    }

    static void show_abs(const char*, const absval&); // for debugging

public:
    integer() : val(), neg(false) { }

    integer(const integer&) = default;
    integer(integer&&) = default;

    integer& operator=(const integer&) = default;
    integer& operator=(integer&&) = default;

    ~integer() noexcept { }

    void clear() noexcept
    {
        val.clear();
        neg = false;
    }

    void swap(integer& other)
    {
        val.swap(other.val);
        std::swap(neg, other.neg);
    }

private:
    //
    // Construct from absolute value:
    //
    integer(const absval& rhs) : val(rhs), neg(false) { }
    integer(absval&& rhs) : val(std::move(rhs)), neg(false) { }
    friend integer gcd(const integer&, const integer&);

    //
    // Helpers for other ctors:
    //
    void init(long long);
    void init(unsigned long long);
    void init(long double);
    #include "bignum_type_traits.inc"

public:
    template<typename T>
    integer(T value, typename std::enable_if<is_sgn<T>::value,int>::type = 0)
      : val()
    {
        init(static_cast<long long>(value));
    }
    template<typename T>
    integer(T value, typename std::enable_if<is_uns<T>::value,int>::type = 0)
      : val(), neg(false) // pathological coupling to init(unsigned long long)
    {                     // which assumes neg is already set
        init(static_cast<unsigned long long>(value));
    }
    template<typename T>
    integer(T value, typename std::enable_if<is_bool<T>::value,int>::type = 0)
      : val(), neg(false)
    {
        if (value)
        {
            val.assign(1, 1);
        }
    }
    template<typename T>
    integer(T value, typename std::enable_if<is_flt<T>::value,int>::type = 0)
      : val()
    {
        init(static_cast<long double>(value));
    }

private:
    void init(const char*, int, const char**);
public:
    explicit integer(const char* value, int rad = 0, const char** end = nullptr)
      : val()
    {
        init(value, rad, end);
    }
    explicit integer(const std::string& value,
                     int rad = 0,
                     std::size_t* term = nullptr)
      : val()
    {
        const char* beg = value.c_str();
        const char* end = nullptr;
        init(beg, rad, &end); // Sets end to a pointer to the char
                              // that terminated the parse (maybe '\0').
        if (term != nullptr)
        {
            *term = static_cast<std::size_t>(end - beg);
        }
    }

    template<typename T>
    typename std::enable_if<std::is_arithmetic<T>::value,integer>::type&
      operator=(T rhs)
    {
        return *this = integer(rhs);
    }
    integer& assign(const char* value, int rad = 0, const char** end = nullptr)
    {
        return *this = integer(value, rad, end);
    }
    integer& assign(const std::string& value,
                    int rad = 0,
                    std::size_t* end = nullptr)
    {
        return *this = integer(value, rad, end);
    }

    bool is_zero() const noexcept
    {
        return val.empty();
    }
    bool is_odd() const noexcept
    {
        return !val.empty() && (val.front() & 1);
    }
    bool is_even() const noexcept
    {
        return !is_odd();
    }
    bool is_neg() const noexcept
    {
        return neg;
    }
    bool is_pos() const noexcept
    {
        return !is_neg() && !is_zero();
    }
    bool is_one() const noexcept
    {
        return val.size() == 1 && val.front() == 1;
    }
    bool is_one(bool n) const noexcept
    {
        return neg == n && is_one();
    }

    explicit operator bool() const noexcept
    {
        return !is_zero();
    }

private:
    static bool fits_in_ull(std::size_t siz) noexcept
    {
        // argument is number of val_type elements
        return siz <= sizeof(unsigned long long) / sizeof(val_type);
    }
    static unsigned long long make_ull(const absval&) noexcept;

    template<typename T>
    T to_number(typename std::enable_if<is_sgn<T>::value,int>::type = 0) const
    {
        if (fits_in_ull(val.size()))
        {
            unsigned long long uval = make_ull(val);
            constexpr T max = std::numeric_limits<T>::max();
            if (!neg && uval <= static_cast<unsigned long long>(max))
            {
                return static_cast<T>(uval);
            }
            if (neg && (uval <= LLONG_MAX ||
                        LLONG_MIN != -LLONG_MAX &&
                        LLONG_MIN == static_cast<long long>(uval)))
            {
                long long sval = uval <= LLONG_MAX ?
                                 -static_cast<long long>(uval) : LLONG_MIN;
                if (sval >= std::numeric_limits<T>::min())
                {
                    return static_cast<T>(sval);
                }
            }
        }
        types::throw_overflow("integer", types::name<T>::value);
    }

    template<typename T>
    T to_number(typename std::enable_if<is_uns<T>::value,int>::type = 0) const
    {
        if (fits_in_ull(val.size()))
        {
            if (!neg) // as expected converting to an unsigned type
            {
                unsigned long long temp = make_ull(val);
                if (temp <= std::numeric_limits<T>::max())
                {
                    return static_cast<T>(temp);
                }
            }
            else // try to modwrap like built-in types
            {
                try
                {
                    return static_cast<T>(
                        to_number<typename std::make_signed<T>::type>());
                }
                catch (const std::overflow_error&)
                {
                    ; // just fall through to exception below
                }
            }
        }
        types::throw_overflow("integer", types::name<T>::value);
    }

    long double make_ld() const noexcept;

    template<typename T>
    T to_number(typename std::enable_if<is_flt<T>::value,int>::type = 0) const
    {
        if (!empty()) // 2023-11-29
        {
            long double temp = make_ld();
            if (std::isfinite(temp) &&
                (std::fabs(temp) <= std::numeric_limits<T>::max()) &&
                (std::fabs(temp) >= std::numeric_limits<T>::min()))
            {
                return static_cast<T>(temp);
            }
            types::throw_overflow("integer", types::name<T>::value);
        }
        return static_cast<T>(0.0);
    }

public:
    explicit operator char() const
    {
        return to_number<char>();
    }
    explicit operator signed char() const
    {
        return to_number<signed char>();
    }
    explicit operator unsigned char() const
    {
        return to_number<unsigned char>();
    }
    explicit operator short() const
    {
        return to_number<short>();
    }
    explicit operator unsigned short() const
    {
        return to_number<unsigned short>();
    }
    explicit operator int() const
    {
        return to_number<int>();
    }
    explicit operator unsigned int() const
    {
        return to_number<unsigned int>();
    }
    explicit operator long() const
    {
        return to_number<long>();
    }
    explicit operator unsigned long() const
    {
        return to_number<unsigned long>();
    }
    explicit operator long long() const
    {
        return to_number<long long>();
    }
    explicit operator unsigned long long() const
    {
        return to_number<unsigned long long>();
    }
    explicit operator float() const
    {
        return to_number<float>();
    }
    explicit operator double() const
    {
        return to_number<double>();
    }
    explicit operator long double() const
    {
        return to_number<long double>();
    }

    std::string to_string(int = 10) const;

    int compare(const integer&) const noexcept;

    int signum() const noexcept
    {
        return neg ? -1 : is_zero() ? 0 : +1;
    }

    integer& set_to_zero() noexcept
    {
        clear();
        return *this;
    }
    integer& set_to_one(bool n = false)
    {
        val.assign(1, 1);
        neg = n;
        return *this;
    }

    integer& negate() noexcept
    {
        if (!is_zero())
        {
            neg = !neg;
        }
        return *this;
    }
    integer& abs() noexcept
    {
        neg = false;
        return *this;
    }

private:
    void do_inc_dec(bool);
public:
    integer& operator++()
    {
        do_inc_dec(false);
        return *this;
    }
    integer& operator--()
    {
        do_inc_dec(true);
        return *this;
    }

    integer operator++(int)
    {
        integer temp(*this);
        do_inc_dec(false);
        return temp;
    }
    integer operator--(int)
    {
        integer temp(*this);
        do_inc_dec(true);
        return temp;
    }

private:
    static int compare(const absval&, const absval&) noexcept;

    static void add(absval&, const absval&);
    static void sub(absval&, const absval&);

    static void mul_part(absval&, val_type);
    static void mul(integer&, const integer&);

    static ext_type div_part(absval&, const absval&, ext_type, ext_type);
    static void do_div(integer*, integer*, const integer&, const integer&);

public:
    integer& operator+=(const integer&);
    integer& operator-=(const integer&);
    integer& operator*=(const integer&);
    integer& operator/=(const integer&);
    integer& operator%=(const integer&);

    struct div_t;
    div_t div(const integer&) const;

    integer& complement() noexcept;
    integer& operator&=(const integer&) noexcept;
    integer& operator^=(const integer&);
    integer& operator|=(const integer&);
private:
    void shift_left(unsigned);
    void shift_right(unsigned);
public:
    integer& operator<<=(int);
    integer& operator>>=(int);

    bool empty() const noexcept
    {
        return val.empty();
    }
    std::size_t elem_size() const noexcept
    {
        return sizeof(val_type);
    }
    std::size_t size() const noexcept
    {
        return val.size() * sizeof(val_type);
    }
    std::size_t capacity() const noexcept
    {
        return val.capacity() * sizeof(val_type);
    }
    void shrink_to_fit()
    {
        val.shrink_to_fit();
    }
};
struct integer::div_t final
{
    integer quot, rem;
    void swap(div_t& other)
    {
        quot.swap(other.quot);
        rem.swap(other.rem);
    }
};
inline integer::div_t integer::div(const integer& divisor) const
{
    div_t retval;
    do_div(&retval.quot, &retval.rem, *this, divisor);
    return retval;
}
inline integer::div_t div(const integer& dividend, const integer& divisor)
{
    integer::div_t retval(std::move(dividend.div(divisor)));
    return retval;
}

using std::swap;
inline void swap(integer& lhs, integer& rhs)
{
    lhs.swap(rhs);
}
inline void swap(integer::div_t& lhs, integer::div_t& rhs)
{
    lhs.swap(rhs);
}

inline integer abs(const integer& val)
{
    integer temp(val);
    temp.abs();
    return temp;
}

inline std::string to_string(const integer& val, int radix = 10)
{
    std::string retval(std::move(val.to_string(radix)));
    return retval;
}

integer gcd(const integer&, const integer&);
integer lcm(const integer&, const integer&);

inline bool operator==(const integer& lhs, const integer& rhs) noexcept
{
    return lhs.compare(rhs) == 0;
}
inline bool operator!=(const integer& lhs, const integer& rhs) noexcept
{
    return lhs.compare(rhs) != 0;
}
inline bool operator<(const integer& lhs, const integer& rhs) noexcept
{
    return lhs.compare(rhs) < 0;
}
inline bool operator>(const integer& lhs, const integer& rhs) noexcept
{
    return lhs.compare(rhs) > 0;
}
inline bool operator<=(const integer& lhs, const integer& rhs) noexcept
{
    return lhs.compare(rhs) <= 0;
}
inline bool operator>=(const integer& lhs, const integer& rhs) noexcept
{
    return lhs.compare(rhs) >= 0;
}

inline integer operator+(const integer& rhs)
{
    return rhs;
}
inline integer operator-(const integer& rhs)
{
    integer temp(rhs);
    temp.negate();
    return temp;
}

inline integer operator+(const integer& lhs, const integer& rhs)
{
    integer temp(lhs);
    temp += rhs;
    return temp;
}
inline integer operator-(const integer& lhs, const integer& rhs)
{
    integer temp(lhs);
    temp -= rhs;
    return temp;
}
inline integer operator*(const integer& lhs, const integer& rhs)
{
    integer temp(lhs);
    temp *= rhs;
    return temp;
}
inline integer operator/(const integer& lhs, const integer& rhs)
{
    integer temp(lhs);
    temp /= rhs;
    return temp;
}
inline integer operator%(const integer& lhs, const integer& rhs)
{
    integer temp(lhs);
    temp %= rhs;
    return temp;
}

inline integer operator~(const integer& rhs)
{
    integer temp(rhs);
    temp.complement();
    return temp;
}
inline integer operator&(const integer& lhs, const integer& rhs)
{
    integer temp(lhs);
    temp &= rhs;
    return temp;
}
inline integer operator^(const integer& lhs, const integer& rhs)
{
    integer temp(lhs);
    temp ^= rhs;
    return temp;
}
inline integer operator|(const integer& lhs, const integer& rhs)
{
    integer temp(lhs);
    temp |= rhs;
    return temp;
}
inline integer operator<<(const integer& lhs, int rhs)
{
    integer temp(lhs);
    temp <<= rhs;
    return temp;
}
inline integer operator>>(const integer& lhs, int rhs)
{
    integer temp(lhs);
    temp >>= rhs;
    return temp;
}

} // namespace bignum

#ifndef INTEGER_NO_IO_OPERATORS
  #include "integer_io.hpp"
#endif

#endif // INTEGER_HPP_INCLUDED
